open Gny;;
open Fwd_rules;;

let is_in_pool pool f =
  try in_pool pool f; true
      with Not_found -> false;;

let rec merge_new old = List.fold_left
  (fun s ((f,_) as fpr) -> match s with
      None ->
	(if is_in_pool old f
	    then None
	    else Some (Gny_prime_map.add f fpr old)
        ) 
    | Some z ->
	(if is_in_pool z f
	    then Some z
	    else Some (Gny_prime_map.add f fpr z)
        )
  ) None;;

let rec fwd_chain
  (progress : unit->unit)
  (old : proof Gny_prime_map.t) =
    match merge_new old (apply_rules old)
    with Some z -> progress (); fwd_chain progress z
       | None -> old;;
 
(*
  | (x, x_proof)::tail ->
      progress ();
      if ( try  Gny_prime_map.find x old; true
           with Not_found -> false
         )
      then fwd_chain progress old tail
      else let old2= Gny_prime_map.add x x_proof old
           in fwd_chain progress old2 (tail @ (apply_rules old2));;
*)
type answer =
    Good of proof
  | Bad of gny list;;

let fwd (hypotheses : (string*gny) list) (conclusion : gny) =
  let init_set =
      List.fold_left
      (fun s (x,y) -> Gny_prime_map.add x (x,y) s) Gny_prime_map.empty 
      (((true, conclusion),(Hypothesis "C"))::
        (List.map (fun (hname,x) -> ((false, x),
        (Hypothesis hname))) hypotheses))
  in
  let reached =
    fwd_chain (function ()->output_char stderr '.'; flush stderr) init_set
  in  try  Good(Gny_prime_map.find (false,conclusion) reached)
      with Not_found ->
             Bad(Gny_prime_map.fold
		   (fun (flag,formula) _ s ->
		     match flag with
		       true when
		         formula <> conclusion &&
		         ( try Gny_prime_map.find (false,formula) reached;
                               false
			   with Not_found -> true
                         ) ->
			   formula::s
		     | _ -> s
		   ) reached []
		);;
(*
let hypotheses = [
  Possesses("P", Encrypt_s((Variable "X"), Pair((Variable "K"), (Variable "K'"))));
  Possesses("P", (Variable "K"));
  Possesses("P", (Variable "K'"))]
and conclusion = Possesses("P", (Variable "X"));;

print_endline (match (fwd hypotheses conclusion)
 with Good _ -> "good" | Bad _ -> "bad");;
*)

