%{
open Protocol_syntax;; open Gny;;
%}

%token IS_TOLD POSSESSES ONCE_CONVEYED BELIEVES FRESH RECOGNIZABLE
%token LEFT_ARROW RIGHT_ARROW VDASH
%token ENCRYPT_LEFT ENCRYPT_RIGHT PLUS MINUS
%token COMMA PAREN_LEFT PAREN_RIGHT DELIMITER SEMICOLON COLON
%token SLASH EXCLAMATION_POINT STAR
%token HONEST_AND_COMPETENT DOUBLE_RIGHT_ARROW CURVY_RIGHT_ARROW
%token PROTOCOL END GOALS MESSAGES ASSUMPTIONS 
%token EOF
%token <string> Identifier String

%start question
%type <Protocol_syntax.question> question
%type <gny> gny gny_nopkey gny_group

%%
gny_nopkey:
    principal IS_TOLD gny { Is_told($1,$3) }
  | principal POSSESSES gny { Possesses($1,$3) }
  | principal ONCE_CONVEYED gny { Once_conveyed($1, $3) }
  | principal BELIEVES gny { Believes($1, $3) }
  | principal DOUBLE_RIGHT_ARROW gny { Jurisdiction($1, $3) }
  | FRESH gny { Fresh($2) }
  | RECOGNIZABLE gny { Recognizable($2) }
  | principal LEFT_ARROW gny RIGHT_ARROW principal { Is_secret($1, $5, $3) }
  | VDASH PLUS pkey_id RIGHT_ARROW principal { Is_public_key($5, $3) }
  | ENCRYPT_LEFT gny_group ENCRYPT_RIGHT PLUS pkey_id { Encrypt_pub($2, $5) }
  | ENCRYPT_LEFT gny_group ENCRYPT_RIGHT MINUS pkey_id { Encrypt_priv($2, $5) }
  | ENCRYPT_LEFT gny_group ENCRYPT_RIGHT gny_nopkey { Encrypt_s($2, $4) }
  | ENCRYPT_LEFT gny_group ENCRYPT_RIGHT EXCLAMATION_POINT gny_nopkey
      { Decrypt_s($2, $5) }
  | function_id PAREN_LEFT gny PAREN_RIGHT { Unary_func($1, $3) }
  | function_id PAREN_LEFT gny COMMA gny PAREN_RIGHT {Binary_func($1, $3, $5)}
  | variable_id { Variable($1) }
  | STAR gny_nopkey { Not_originated_here($2) }
  | PAREN_LEFT gny_group PAREN_RIGHT { $2 }
  | principal HONEST_AND_COMPETENT { Honest_and_competent($1) }
  | PAREN_LEFT gny CURVY_RIGHT_ARROW gny PAREN_RIGHT
      { Leading_to($2, $4) }

gny:
    gny_nopkey { $1 }
  | PLUS pkey_id { Public_key($2) }
  | MINUS pkey_id { Private_key($2) }
  
gny_group:
    gny { $1 }
  | gny COMMA gny_group { Pair($1, $3) }

tformula:
    formula_id COLON gny SEMICOLON { $1, $3 }

tformulae:
    { [] }
  | tformula tformulae { $1::$2 }

formula:
    gny SEMICOLON { $1 }

formulae:
    { [] }
  | formula formulae { $1::$2 }

message:
    principal RIGHT_ARROW principal COLON gny SEMICOLON
      { { source = $1;
	  destination = $3;
	  message = $5
	} }

messages:
    { [] }
  | message messages { $1::$2 }

declarations:
      { () }

question:
    PROTOCOL protocol_id SEMICOLON declarations 
    ASSUMPTIONS tformulae
    MESSAGES messages
    GOALS formulae
    END EOF { {protocol_name =  $2; assumptions = $6;
               messages = $8; goals=$10 } }

principal:
    Identifier { $1 }

pkey_id:
    Identifier { $1 }

function_id:
    Identifier { $1 }

variable_id:
    Identifier { $1 }

formula_id:
    Identifier { $1 }

protocol_id:
    String { $1 }


