open Rules_syntax;;
open Gny;;
open Gny_latex;;

let rules_data = open_in "gny_rules.dat.txt";;
let ml_code = open_out "fwd_rules.ml";;
let rules_latex = open_out "fwd_rules.tex";;

let description = Rules_parser.rules Rules_lexer.rules_lex
  (Lexing.from_channel rules_data);;

let latex_stuff2 l =
  String.concat " \\rulesep "
    (List.map gny_to_latex l);;

let latex_stuff a b =
  let a'=latex_stuff2 a and b'=latex_stuff2 b in
  if a'=""
  then b'
  else (Printf.sprintf "\\fbox{$%s$}" a') ^
    (if b'=""
    then ""
    else " \\rulesep " ^ b');;

let latex_rule a b c d =
  Printf.sprintf "\\frac{%s}{%s}" (latex_stuff a b) (latex_stuff c d);;

List.iter (function id, z ->
  Printf.fprintf rules_latex "\\item \\textsf{%s} is %s rule:
$$%s$$
" id (match z with
    Construct _ -> "a construction"
  | Destruct _ -> "a destruction"
  | Equivalence _ -> "an equivalence")
    (match z with
      Construct(h, c) -> latex_rule [] h [] c
    | Destruct(h, k, c) -> latex_rule h k [] c
    | Equivalence(h, c) -> latex_rule [h] [] [] [c]) ) description;;

let gny_prime_rules rules =
  List.flatten
  ( List.map
      (function id, z -> match z with
          Construct(h, c) ->
            List.flatten
            ( List.map
               (function c ->
                 [
                   { name = id;
                     matching_hypotheses = [true,c];                   
                     other_hypotheses = (List.map (function f -> false, f) h);
                     conclusions = [false, c]
                   };
                   { name = id ^ "b";
		     matching_hypotheses = [true, c];
		     other_hypotheses = [];
		     conclusions = (List.map (function f -> true, f) h)
		   } 
                 ] 
               ) c
            )
	| Destruct(h1, h2, c) ->
	    { name =id;
                matching_hypotheses =
                   (List.map (function f -> false, f) h1);
                other_hypotheses =
                   (List.map (function f -> false, f) h2);
                conclusions =
                   (List.map (function f -> false, f) c)
            } ::
            ( match h1 with
	        [] -> []
              |  _ ->
	      [{ name=id ^ "t";
	        matching_hypotheses =
                   (List.map (function f -> false, f) h1);
	        other_hypotheses = [];
		conclusions = (List.map (function f -> true, f) h2)
	      }]
            )
        | Equivalence(h, c) ->
            [ { name =id;
                matching_hypotheses = [false, h];
                other_hypotheses    = [];
                conclusions         = [false, c]
              }
            ]	
     ) rules
  );;

output_string ml_code
  (Rules_ml_code.generate_ml_rules_application (gny_prime_rules description));;

close_in rules_data;;
close_out ml_code;;
close_out rules_latex;;

