open Gny;;

module String_map = Map.Make
  ( struct
      type t=string
      let compare=Pervasives.compare
    end
  );;

type variable_type = Principal | Pkey_id | Formula
                   | Unary_function | Binary_function;;

let variable_type_to_string = function
    Principal -> "principal"
  | Pkey_id -> "public-key/private-key pair id"
  | Formula -> "formula"
  | Unary_function -> "unary_function"
  | Binary_function -> "binary_function";;

(*
let seen_variable var type1 map =
  let type2, count=
      try String_map.find var map
      with Not_found -> type1, 0
  in  ( if type1<>type2

        then failwith (Printf.sprintf "%s is used with types %s and %s"
                         var (variable_type_to_string type1)
                             (variable_type_to_string type2))
      );
      String_map.add var (type1, (succ count)) map;;

let rec check_variables vars = function
    Is_told(p, x) -> seen_variable p Principal (check_variables vars x)
  | Possesses(p, x) -> seen_variable p Principal (check_variables vars x)
  | Once_conveyed(p, x) -> seen_variable p Principal (check_variables vars x)
  | Believes(p, x) -> seen_variable p Principal (check_variables vars x)
  | Fresh(x) -> check_variables vars x
  | Recognizable(x) -> check_variables vars x
  | Is_secret(p, q, x) -> seen_variable p Principal
                        ( seen_variable q Principal
                        ( check_variables vars x) )
  | Is_public_key(p, k) -> seen_variable p Principal
                          (seen_variable k Pkey_id vars)
  | Encrypt_pub(x, k) -> seen_variable k Pkey_id (check_variables vars x) 
  | Encrypt_priv(x, k) -> seen_variable k Pkey_id (check_variables vars x) 
  | Encrypt_s(x, k) -> check_variables (check_variables vars k) x
  | Pair(x, k) -> check_variables (check_variables vars k) x
  | Unary_func(f, x) -> seen_variable f Unary_function (check_variables vars x)  | Binary_func(f, x, y) -> seen_variable f Unary_function
                           (check_variables (check_variables vars y) x)
  | Variable(x) -> seen_variable x Formula vars
  | Public_key(k) -> seen_variable k Pkey_id vars
  | Private_key(k) -> seen_variable k Pkey_id vars;;
*)

type rule_variables = (variable_type*int)String_map.t;;

let check_in_ml_variable var type1 (map : rule_variables) =
  let var=String.lowercase var in
  let var'="z_" ^ var in
  let type2, count=
      try String_map.find var' map
      with Not_found -> type1, 0
  in  ( if type1<>type2
        then failwith (Printf.sprintf "%s is used with types %s and %s"
                         var (variable_type_to_string type1)
                             (variable_type_to_string type2))
      );
      (if count=0
       then var'
       else Printf.sprintf "%s_%d" var' (succ count)),
      (String_map.add var' (type1, (succ count)) map);;

let rec generate_ml_matching (vars : rule_variables) = function
    Is_told(p, x) -> let x', vars2 = generate_ml_matching vars x in
                     let p', vars3 = check_in_ml_variable p Principal vars2 in
                     (Printf.sprintf "Is_told(%s, %s)" p' x'), vars3
  | Possesses(p, x) -> let x', vars2 = generate_ml_matching vars x in
                     let p', vars3 = check_in_ml_variable p Principal vars2 in
                     (Printf.sprintf "Possesses(%s, %s)" p' x'), vars3
  | Once_conveyed(p, x) -> let x', vars2 = generate_ml_matching vars x in
                     let p', vars3 = check_in_ml_variable p Principal vars2 in
                     (Printf.sprintf "Once_conveyed(%s, %s)" p' x'), vars3
  | Believes(p, x) -> let x', vars2 = generate_ml_matching vars x in
                     let p', vars3 = check_in_ml_variable p Principal vars2 in
                     (Printf.sprintf "Believes(%s, %s)" p' x'), vars3
  | Fresh(x) -> let x', vars2 = generate_ml_matching vars x in
                (Printf.sprintf "Fresh(%s)" x'), vars2
  | Recognizable(x) -> let x', vars2 = generate_ml_matching vars x in
                (Printf.sprintf "Recognizable(%s)" x'), vars2
  | Is_secret(p, q, x) -> let x', vars2 = generate_ml_matching vars x in
                     let p', vars3 = check_in_ml_variable p Principal vars2 in
                     let q', vars4 = check_in_ml_variable q Principal vars3 in
                     (Printf.sprintf "Is_secret(%s, %s, %s)" p' q' x'), vars4
  | Is_public_key(p, k) ->
                     let k', vars2 = check_in_ml_variable k Pkey_id vars in
                     let p', vars3 = check_in_ml_variable p Principal vars2 in
                     (Printf.sprintf "Is_public_key(%s, %s)" p' k'), vars3
  | Encrypt_pub(x, k) ->
                     let k', vars2 = check_in_ml_variable k Pkey_id vars in
		     let x', vars3 = generate_ml_matching vars2 x in
		     (Printf.sprintf "Encrypt_pub(%s, %s)" x' k'), vars3
  | Encrypt_priv(x, k) ->
                     let k', vars2 = check_in_ml_variable k Pkey_id vars in
		     let x', vars3 = generate_ml_matching vars2 x in
		     (Printf.sprintf "Encrypt_priv(%s, %s)" x' k'), vars3
  | Encrypt_s(x, k) ->
                     let k', vars2 = generate_ml_matching vars k in                                 let x', vars3 = generate_ml_matching vars2 x in
		     (Printf.sprintf "Encrypt_s(%s, %s)" x' k'), vars3
  | Decrypt_s(x, k) ->
                     let k', vars2 = generate_ml_matching vars k in                                 let x', vars3 = generate_ml_matching vars2 x in
		     (Printf.sprintf "Decrypt_s(%s, %s)" x' k'), vars3
  | Unary_func(f, x) ->
                     let x', vars2 = generate_ml_matching vars x in
		     let f', vars3 = check_in_ml_variable f Unary_function vars2 in
		     (Printf.sprintf "Unary_func(%s,%s)" f' x'), vars3
  | Binary_func(f, x, y) ->
                     let x', vars2 = generate_ml_matching vars x in
                     let y', vars3 = generate_ml_matching vars2 y in
		     let f', vars4 = check_in_ml_variable f Binary_function vars3 in
		     (Printf.sprintf "Binary_func(%s,%s, %s)" f' x' y'), vars4
  | Variable(x) -> check_in_ml_variable x Formula vars
  | Public_key(k) -> let k', vars2=check_in_ml_variable k Pkey_id vars
                     in (Printf.sprintf "Public_key(%s)" k'), vars2
  | Private_key(k) -> let k', vars2=check_in_ml_variable k Pkey_id vars
                     in (Printf.sprintf "Private_key(%s)" k'), vars2
  | Pair(x, y) ->
                     let y', vars2 = generate_ml_matching vars y in                                 let x', vars3 = generate_ml_matching vars2 x in
		     (Printf.sprintf "Pair(%s, %s)" x' y'), vars3
  | Not_originated_here(x) ->
      let x', vars2 = generate_ml_matching vars x in
      (Printf.sprintf "Not_originated_here(%s)" x'), vars2
  | Honest_and_competent(p) ->
      let p', vars2 = check_in_ml_variable p Principal vars in
      (Printf.sprintf "Honest_and_competent(%s)" p'), vars2
  | Jurisdiction(p, x) ->
      let p', vars2 = check_in_ml_variable p Principal vars in
      let x', vars3 = generate_ml_matching vars2 x in
      (Printf.sprintf "Jurisdiction(%s,%s)" p' x'), vars3
  | Leading_to(x, y) ->
                     let y', vars2 = generate_ml_matching vars y in                                 let x', vars3 = generate_ml_matching vars2 x in
		     (Printf.sprintf "Leading_to(%s, %s)" x' y'), vars3;;

let matching_constraints
  (precedent : rule_variables)
  (current   : rule_variables) =
  String_map.fold
    ( fun (var : string) (_, (current_count : int)) (s : string) ->
        if current_count>1
        then ( let precedent_count =
                 try snd(String_map.find var precedent)
                 with Not_found -> 1
               in  let rec stuff s n =
                     if n<=current_count
                     then stuff
                          (
                            ( match s with
                                "" -> " when "
                              | _  -> s ^ " && "
                            ) ^ var ^ "=" ^ var ^ "_" ^ (string_of_int n)
                          ) (succ n)
                     else s
                   in stuff s (succ precedent_count)
             )
        else s 
    ) current "";;

let bool_to_string = function true -> "true" | false -> "false";;

let generate_ml_match_clause
  (precedent    : rule_variables)
  (flag,formula : gny_prime) =
  let formula', current = generate_ml_matching precedent formula
  in ((bool_to_string flag) ^ ", (" ^ formula' ^ ")"
       ^ (matching_constraints precedent current)
     ), current;;

let rec generate_other_matching_tests
  (follow_vars : rule_variables)
  (prepend : string)
  (middle : string->rule_variables->string)
 (index  : int)
  = function
      [] -> (middle prepend follow_vars)
    | first::others ->
        let fm, next_vars = generate_ml_match_clause follow_vars first in
          prepend ^ "walk_pool old\n" ^ prepend
        ^ "  (function proof" ^ (string_of_int index) ^ " -> function\n"
        ^ prepend ^ "      " ^ fm ^ " ->\n"
        ^ (generate_other_matching_tests next_vars
             (prepend ^ "        ") middle (succ index) others)
        ^ prepend ^ "    | _ -> []\n" ^ prepend ^ "  )\n";;

let check_variable vars var typ1 =
  let var = String.lowercase var in
  let var' = "z_" ^ var in
  let typ2,_ =
        try String_map.find var' vars
        with Not_found->failwith ("Variable " ^ var ^ " unknown.")
  in if typ1<>typ2
     then failwith (Printf.sprintf "%s is used with types %s and %s"
                         var (variable_type_to_string typ1)
                             (variable_type_to_string typ2))
     else var';;       

let rec generate_ml_formula (vars : rule_variables) = function
    Is_told(p, x) -> let p' = check_variable vars p Principal
                     and x' = generate_ml_formula vars x in
                     (Printf.sprintf "Is_told(%s, %s)" p' x')
  | Possesses(p, x) -> let p' = check_variable vars p Principal
                     and x' = generate_ml_formula vars x in
                     (Printf.sprintf "Possesses(%s, %s)" p' x')
  | Once_conveyed(p, x) -> let p' = check_variable vars p Principal
                     and x' = generate_ml_formula vars x in
                     (Printf.sprintf "Once_conveyed(%s, %s)" p' x')
  | Believes(p, x) -> let p' = check_variable vars p Principal
                     and x' = generate_ml_formula vars x in
                     (Printf.sprintf "Believes(%s, %s)" p' x')
  | Fresh(x) ->  let x' = generate_ml_formula vars x in
                (Printf.sprintf "Fresh(%s)" x')
  | Recognizable(x) ->let x' = generate_ml_formula vars x in
                (Printf.sprintf "Recognizable(%s)" x')
  | Is_secret(p, q, x) -> let p' = check_variable vars p Principal
                     and q' = check_variable vars q Principal
                     and x' = generate_ml_formula vars x in
                     (Printf.sprintf "Is_secret(%s, %s, %s)" p' q' x')
  | Is_public_key(p, k) -> let p' = check_variable vars p Principal
                           and k' = check_variable vars k Pkey_id in
                     (Printf.sprintf "Is_public_key(%s, %s)" p' k')
  | Encrypt_pub(x, k) -> let x'=generate_ml_formula vars x
                           and k' = check_variable vars k Pkey_id in
 		     (Printf.sprintf "Encrypt_pub(%s, %s)" x' k')
  | Encrypt_priv(x, k) -> let x'=generate_ml_formula vars x
                           and k' = check_variable vars k Pkey_id in
		     (Printf.sprintf "Encrypt_priv(%s, %s)" x' k')
  | Encrypt_s(x, k) -> let x' = generate_ml_formula vars x
                       and k' = generate_ml_formula vars k in
		     (Printf.sprintf "Encrypt_s(%s, %s)" x' k')
  | Decrypt_s(x, k) -> let x' = generate_ml_formula vars x
                       and k' = generate_ml_formula vars k in
		     (Printf.sprintf "Decrypt_s(%s, %s)" x' k')
  | Unary_func(f, x) -> let f' = check_variable vars f Unary_function
                        and x' = generate_ml_formula vars x in
		     (Printf.sprintf "Unary_func(%s,%s)" f' x')
  | Binary_func(f, x, y) -> let f' = check_variable vars f Binary_function
                        and x' = generate_ml_formula vars x
                        and y' = generate_ml_formula vars y in
		     (Printf.sprintf "Binary_func(%s,%s, %s)" f' x' y')
  | Variable(x) -> check_variable vars x Formula
  | Public_key(k) -> Printf.sprintf "Public_key(%s)"
                       (check_variable vars k Pkey_id)
  | Private_key(k) -> Printf.sprintf "Private_key(%s)"
                       (check_variable vars k Pkey_id)
  | Pair(x, y) -> let x' = generate_ml_formula vars x
                  and y' = generate_ml_formula vars y
		  in (Printf.sprintf "Pair(%s, %s)" x' y')
  | Not_originated_here(x) ->
      let x' = generate_ml_formula vars x
      in Printf.sprintf "Not_originated_here(%s)" x'
  | Honest_and_competent(p) ->
      Printf.sprintf "Honest_and_competent(%s)"
        (check_variable vars p Principal)
  | Jurisdiction(p, x) ->
      Printf.sprintf "Jurisdiction(%s, %s)"
	(check_variable vars p Principal)
	(generate_ml_formula vars x)
  | Leading_to(x, y) -> let x' = generate_ml_formula vars x
                  and y' = generate_ml_formula vars y
		  in (Printf.sprintf "Leading_to(%s, %s)" x' y')

let generate_ml_formula_prime vars (flag, formula) =
  "(" ^ (bool_to_string flag) ^ ","
             ^ (generate_ml_formula vars formula) ^ ")";;

let generate_other_checks
  (other_hypotheses : gny_prime list)
  (vars             : rule_variables)
  (prepend          : string)
  (middle           : string->rule_variables->string)
  (start_index      : int) =
  ( if   other_hypotheses = []
    then middle prepend vars
    else prepend ^ "( try  let " ^
      (fst (List.fold_left
        (fun (s,index) formula ->
           ( ( match s with
                 "" -> ""
               | _ -> s ^ "\n" ^ prepend ^ "       and "
             ) ^ "proof" ^ (string_of_int index) ^ " = in_pool old "
             ^ (generate_ml_formula_prime vars formula)
           ), (succ index)
        ) ("",start_index) other_hypotheses
      ))^ " in\n" ^ (middle (prepend ^ "  ") vars)
      ^ prepend ^ "  with Not_found -> []\n" ^ prepend ^ ")\n"
  );;

let rec proofs = function
    0 -> ""
  | 1 -> "proof1"
  | n -> proofs (pred n) ^ "; proof" ^ (string_of_int n);;

let generate_first_test
  { name                = name;
    matching_hypotheses = matching_hypotheses;
    other_hypotheses    = other_hypotheses;
    conclusions         = conclusions } =
  match matching_hypotheses with
    [] -> ""
  | first::others ->
      let fm, follow_vars = generate_ml_match_clause String_map.empty first
      in "    (* rule " ^ name ^ " *)\n    ( match f with\n        "
       ^ fm ^ " ->\n"
       ^ (generate_other_matching_tests follow_vars "          "
            (let shift=2+(List.length others)
             and conclude prepend vars =
                 prepend ^ "[" ^
                 ( let proof=
                      Printf.sprintf "Rule(\"%s\", [%s])"
                      name (proofs ((List.length matching_hypotheses)
                                   +(List.length other_hypotheses)))
                   in
                   String.concat (";\n" ^ prepend)
                   (List.map (fun x ->
                      Printf.sprintf "(%s, %s)"
                      (generate_ml_formula_prime vars x) proof
                             )
                      conclusions)
                 ) ^ "]\n" in
             fun prepend vars ->
               generate_other_checks other_hypotheses
                 vars prepend conclude shift) 2 others)
       ^ "      | _ -> []\n    )\n";;

let generate_ml_rules_application
  (rules : gny_prime_rule list) =
  "(* automatically generated *)

open Gny;;

module Gny_prime_map = Map.Make
  ( struct type t=gny_prime
           let  compare=Pervasives.compare
    end
  );;

let walk_pool (pool : proof Gny_prime_map.t)
  (g : proof->gny_prime->proof list) =
  Gny_prime_map.fold
  ( fun x x_proof s ->
      try s @ (g x_proof x)
      with Not_found -> s
  ) pool [];;

let in_pool
    (pool : proof Gny_prime_map.t)
    (x : gny_prime) =
    (Gny_prime_map.find x pool);;

let rec apply_rules_r (old : proof Gny_prime_map.t)
  (proof1 : proof) (f : gny_prime) =
  (
" ^
  ( String.concat "    @\n\n"
    ( List.map generate_first_test rules
    )
  ) ^ "
    @

    (* rationality rule*)
    ( match f with
        flag, (Believes(p, x)) ->
          List.map (fun ((flag,x),proof) -> (flag, Believes(p, x)), proof)
            (apply_rules_r
              (Gny_prime_map.fold
                (fun (flag,f) pr s ->
		  match f with
                    Believes(p2, x) when p2=p ->
                      Gny_prime_map.add (flag,x) pr s
                  | _ -> s
                ) old
              Gny_prime_map.empty) proof1 (flag,x))
      | _ -> []
    )
  );;

let apply_rules (old : proof Gny_prime_map.t) =
  walk_pool old (apply_rules_r old);;
";;






